"""
Frequently-used modules like self-attention modules of vanilla Transformer are put in this package.
"""

# Created by Wenjie Du <wenjay.du@gmail.com>
# License: BSD-3-Clause

from .base_model_core import ModelCore

__all__ = [
    "ModelCore",
]
