"""
This module contains the configuration for the dataset module.
"""

# Created by Wenjie Du <wenjay.du@gmail.com>
# License: BSD-3-Clause

# Currently we only support h5 files
SUPPORTED_DATASET_FILE_FORMATS = [
    "hdf5",
]
